// Requires java SE 8
// Compilation: javac Human.java
// Execution: java Human

// Java 8 time API
import java.time.Year;

// #### NOTE: there's a way to use integers as unsigned values in Java SE 8,
// #### but it's just freaking boring, so we stick to simple int's.

public class Human
{
  public static void census ()
  {
    if (population != 0)
      System.out.println (population
			  + " human" + ((population > 1) ? "s " : " ")
			  + "currently alive.");
    else
      System.out.println ("No human currently alive.");
  }

  public Human (String _name, float _size, int _birthYear)
  {
    name = _name;
    size = _size;
    birthYear = _birthYear;
    System.out.println ("Creating the human " + name + "...");
    population += 1;
  }
  // #### NOTE: we can't predict when exactly this method is called (GC time),
  // #### so most of the time, the population's count will be
  // #### inaccurate. People usually prefer to implement a custom cleanup
  // #### method such as die().
  // #### WARNING: in fact, this API has been deprecated in Java 9, in favor
  // #### of so-called Cleaners.
  public void finalize () throws Throwable
  {
    try
      {
	System.out.println ("GC'ing the human " + name + "...");
	population -= 1;
      }
    finally
      {
	super.finalize ();
      }
  }

  public String name ()
  {
    return name;
  }
  public void rename (String _name)
  {
    name = _name;
  }

  public int age ()
  {
    return Year.now ().getValue () - birthYear;
  }
  public void hello ()
  {
    System.out.println ("Hello! I'm " + name + ", "
			+ size + "m, "
			+ age () + "yo.");
  }

  private static int population = 0;

  private String name;
  private final float size;
  private final int birthYear;

  // There's no friend concept in Java. There are workarounds though (Cf. the
  // Friend Accessor pattern and other tricks such as using protected members
  // and different classes in the same package, Cf. chapter 2).
  public static void main (String[] args)
  {
    Human.census ();
    System.out.println ();

    // Contrary to C++, you don't get to stack-allocate objects in Java. Also,
    // you don't delete explicitly since the language has a garbage
    // collector.
    Human h1 = new Human ("Alain Térieur", 1.80f, 1970);
    h1.census ();
    h1.hello ();
    System.out.println ();

    Human h2 = new Human ("Alex Térieur", 1.78f, 1975);
    h2.census ();
    h2.hello ();
    System.out.println ();

    Human h3 = new Human ("Vladimir Guez", 1.85f, 1980);
    h3.census ();
    System.out.println (h3.name () + " wants a new name...");
    h3.rename ("Thierry Chmonfiss");
    h3.hello ();
    System.out.println ();

    // #### NOTE: you never want to do this for real, as memory management is
    // #### automatic. This is just to force the program into calling one
    // #### finalize() method.
    h3 = null;
    System.gc ();
    // Even here, the population count will most likely be inaccurate.
    Human.census ();
    System.out.println ();
  }
}
